from fileformats.core import __version__
from .basic import Foo, Bar, Baz, Qux
from .headers import (
    YFile,
    ZFile,
    Xyz,
    MyFormat,
    MyFormatGz,
    MyFormatGzX,
    MyFormatX,
    MyBinaryFormat,
    MyHeader,
    MyBinaryFormatX,
    MyOtherBinaryFormatX,
    YourFormat,
    SeparateHeader,
    ImageWithHeader,
    EncodedText,
)
from .classifiers.file import A, B, C, D, E, F, G, H, J, K, L, M, N, P, Q, R, TestField
from .classifiers.generic import Classified, U, V, W, X, Y, Z
from .magic import Magic, MagicVersion

__all__ = [
    "__version__",
    "Foo",
    "Bar",
    "Baz",
    "Qux",
    "YFile",
    "ZFile",
    "Xyz",
    "MyFormat",
    "MyFormatGz",
    "MyFormatGzX",
    "MyFormatX",
    "MyBinaryFormat",
    "MyHeader",
    "MyBinaryFormatX",
    "MyOtherBinaryFormatX",
    "YourFormat",
    "SeparateHeader",
    "ImageWithHeader",
    "EncodedText",
    "A",
    "B",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "J",
    "K",
    "L",
    "M",
    "N",
    "P",
    "Q",
    "R",
    "TestField",
    "Classified",
    "U",
    "V",
    "W",
    "X",
    "Y",
    "Z",
    "Magic",
    "MagicVersion",
]
